/* Add space between collapsible details HTML tags */
details {
    margin-bottom: 1em;
}

/* Darker pygment highlighing of console input/output */
.highlight .go {
    color: #404040;
}

/* White captions in sidebar */
.wy-nav-side p.caption {
    color: #f5f5f5;
}

/* labscript blue, alpha = 83% */
.wy-side-nav-search {
    background: #2946bbd3;
}

.wy-nav-top {
    background: #2946bbd3;
}

/* labscript green, alpha = 75% */
.rst-content .note .admonition-title {
    background: #00804fbf;
}

/* labscript green, alpha = 25% */
.rst-content .note {
    background: #00804f3f;
}

/* labscript red, alpha = 75% */
.rst-content .warning .admonition-title {
    background: #bc294cbf
}

/* labscript red, alpha = 25% */
.rst-content .warning {
    background: #bc294c3b;
}

/* Elevation
*
* Style box-shadows using Material Design's idea of elevation. These particular numbers are taken from here:
*
* https://github.com/material-components/material-components-web
* https://material-components-web.appspot.com/elevation.html
*/

.rst-content img.screenshot {
    border: none;
    /* MD Elevation 8 */
    box-shadow: 0px 5px 5px -3px rgba(0, 0, 0, 0.2),
        0px 8px 10px 1px rgba(0, 0, 0, 0.14), 0px 3px 14px 2px rgba(0, 0, 0, 0.12);
    margin-bottom: 24px;
}

img.labscript-suite-icon {
    min-width: 32px;
}